#ifndef LITTLE_H
#define LITTLE_H

/*
 * File: little.h
 * Author: H.Schulz
 *
 * (C) by Thalion Graphic Solutions eK. 2002, all rights reserved.
 *
 * The Little VM application API:
 * this file declares functions and data structures part of the
 * Little VM public interface, which provide atomic capabilities
 * of accessing objects, fields and invoking methods.
 */

#include <little/ltl_base.h>
#include <little/ltl_ins.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * VM builtin types.
 */

#define LTL_TYPE_VOID		0
#define LTL_PACKAGE_CORE	1
#define LTL_TYPE_CLASS		2
#define LTL_TYPE_METHOD		3
#define LTL_TYPE_ATTR		4
#define LTL_TYPE_REF		5
#define LTL_TYPE_WORD		6
#define LTL_TYPE_SYMBOL		7
#define LTL_TYPE_PSZ		8	/* char * - string table reference */
#define LTL_TYPE_ERROR		9	/* Error symbol thrown by the ltl_threadenv_error_*() methods */


/*
 * Modifiers known by the VM.
 */
enum ltl_modifiers {
	LTL_MOD_FINAL		= 0x01,
	LTL_MOD_NATIVE		= 0x02,
	LTL_MOD_HIDDEN		= 0x04,
	LTL_MOD_STICKY		= 0x08,
	LTL_MOD_CONST		= 0x10,
	LTL_MOD_LOCAL		= 0x20,
	LTL_MOD_ABSTRACT	= 0x40
};

/*
 * Symbols referring memory objects (classes, features).
 */
typedef ltl_key_t	ltl_sym_t;

/*
 * The central data structure: the object.
 * An object has got a type field, which provides information
 * on it's features: knowledge of available methods and attributes
 * stored in the 'fields' field.
 * The 'fields' field may also be used as a value immediately, as in
 * LTL_WORD, or as a reference to some external information, as in
 * opaque classes. 
 */
typedef struct ltl_object {

	ltl_sym_t		type;
	int32_t			nfields;
	struct ltl_object	*fields;

} ltl_object_t;

#define LTL_OBJECT_TYPE(o) ((o)->type)
#define LTL_OBJECT_NFIELDS(o) ((o)->nfields)
#define LTL_OBJECT_FIELD(o,i) (((ltl_object_t *)(o))->fields+(i)) 

/*
 * This allows to initialize 'special' objects, i.e. objects without
 * associated fields, but representing one single value, like integers,
 * symbols, or handles to something.
 * The macro arguments are:
 * (1) pointer to object to set
 * (2) type symbol
 * (3) value - must be castable to pointer without loss
 */
#define LTL_OBJECT_SETSPECIAL(o,t,v) \
	{ (o)->type = t; (o)->nfields = 0; (o)->fields = (ltl_object_t *) (v); }

/*
 * The NIL object.
 */
extern ltl_object_t ltl_nil;
#define LTL_NIL (&ltl_nil);

/*
 * Flags for specifying the Vm verbosity level.
 */
typedef enum {

	LTL_VERBOSE_NO,				/* Silent operation.		*/
	LTL_VERBOSE_CLASS	= 0x1,		/* Be verbose on class loading
						   and initialization.		*/
	LTL_VERBOSE_NATIVELIBS	= 0x2,		/* Be talkative on shared
						   library access.		*/
	LTL_VERBOSE_ERRORS	= 0x4,		/* Extended error info.		*/
	LTL_VERBOSE_THREADS	= 0x8,		/* Threads forking info.	*/
	LTL_VERBOSE_LOOKUPS	= 0x10,		/* Feature lookup info.		*/
	LTL_VERBOSE_JIT		= 0x20		/* Talkative JIT compiler. */
} ltl_verbosity_t;

/*
 * The VM environment:
 * each Little-capable application needs to create at least one
 * ltl_env data structure. The structure contains the global symbol
 * table, and the features data base.
 */
struct ltl_env;
typedef struct ltl_env ltl_env_t;

/*
 * The VM thread environment:
 * execution is always performed within the context of a ltl_threadenv
 * data structure, which contains all required stacks.
 * Depending on availability of threads support, the threadenv may hold
 * an association to an OS thread as well - in this case, execution will
 * take place within this thread's context (custom applications should check
 * for thread support via ltl_env_supports_feature(), if they have to optimally
 * support both threaded and threadless environments).
 * However, it is always correct to start execution via ltl_threadenv_start_and_wait().
 */
struct ltl_threadenv;
typedef struct ltl_threadenv ltl_threadenv_t;

/*
 * Create the ltl_env data structure.
 */
ltl_env_t *
	ltl_env_create(
		int,
		int,
		size_t,
		size_t,
		const char *,
		ltl_verbosity_t			/* Verbosity level flag.	*/
	);

/*
 * Dispose the ltl_env data structure.
 */
void
	ltl_env_dispose(
		ltl_env_t *				/* Little environment.		*/
	);

/*
 * Returns a token for a given string.
 */
ltl_sym_t 
	ltl_env_getsym(
		ltl_env_t *,			/* Little environment.		*/
		const char *			/* The symbol.				*/
	);

/*
 * Returns the string literal referred by a given symbol.
 */
const char *
	ltl_env_get_str4sym(
		ltl_env_t *,			/* Little environment.		*/
		ltl_sym_t				/* Symbol.					*/
	);

/*
 * Tests by symbolic key, if this VM supports a given extension feature
 * (returns true, if yes, else false).
 */
int
	ltl_env_supports_feature(
		ltl_env_t *,			/* The environment.			*/
		ltl_sym_t				/* The extension key symbol.*/
	);

/*
 * Sets an arbitrary on-error callback function.
 * The callback function receives as arguments:
 * - a ltl_threadenv_t object reference, which specifies the error's context
 * - a symbolic error token
 * - three descriptive error strings
 */
void
	ltl_env_set_errorcb1(
		ltl_env_t *,
		void (*)(ltl_threadenv_t *,ltl_sym_t,const char *,const char *,const char *)
	);

/*
 * Sets an arbitrary on-error callback function.
 * The callback function receives as arguments:
 * - a ltl_threadenv_t object reference, which specifies the error's context
 * - a symbolic error token
 * - the current LITTLE receiver object 'self'
 * - the current LITTLE method in execution
 * - the source file name, in which the offending code snippet has been defined
 * - the source line number
 */
void
	ltl_env_set_errorcb2(
		ltl_env_t *,
		void (*)(ltl_threadenv_t *,ltl_sym_t,ltl_object_t *,ltl_object_t *,const char *,int)
	);

void
	ltl_env_set_perror(
		ltl_env_t *,
		void (*)(ltl_threadenv_t *,FILE *)
	);

void
	ltl_env_set_debughook(
		ltl_env_t *,
		void (*)(ltl_threadenv_t *,ltl_object_t *,ltl_object_t *,ltl_object_t *,ltl_ins_t *)
	);

/*
 * Stops the VM (all threads currently in execution).
 */
void
	ltl_env_stopall(
		ltl_env_t *				/* The Little environment.	*/
	);

/*
 * Returns a string containing version and manufacturer info for this VM.
 */
const char *
	ltl_env_get_infostring(
		ltl_env_t *,			/* The Little environment.	*/
		ltl_sym_t
	);

/*
 * For debugging purposes: prints an object structure.
 */
void
	ltl_env_print_object(
		ltl_env_t *,			/* The environment.		*/
		ltl_object_t *,			/* The object to print.		*/
		FILE *					/* File to print to.		*/
	);

/*
 * Creates a thread environment.
 * This environment can then be used to execute Little code.
 */
ltl_threadenv_t *
	ltl_threadenv_create(
		ltl_env_t *		/* ALICE environment.	*/
	);

/*
 * Sets the thread environments object and method - afterwards, the
 * thread execution may be launched via ltl_threadenv_start()
 * or ltl_threadenv_start_and_wait().
 */
ltl_threadenv_t *
	ltl_threadenv_set(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_object_t *,		/* Object.			*/
		ltl_sym_t,		/* Method symbol.		*/
		ltl_object_t *		/* ARGV array, or NULL.		*/
	);

/*
 * Sets the thread environment's run argument.
 */
void
	ltl_threadenv_setarg(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_object_t *		/* Argument object.		*/
	);

/*
 * Starts execution in this thread environment.
 */
int
	ltl_threadenv_start(
		ltl_threadenv_t *
	);
	
/*
 * Starts execution in this thread environment, and waits for the
 * thread to complete.
 */
int
	ltl_threadenv_start_and_wait(
		ltl_threadenv_t *
	);

/*
 * Stops execution in this thread environment.
 */
int
	ltl_threadenv_stop(
		ltl_threadenv_t *
	);

#ifdef LTL_THREADS

# define ltl_threadenv_activate(threadenv)
# define ltl_threadenv_deactivate(threadenv)

#else

/*
 * Marks a thread environment as active (allows proper exception
 * handling - does nothing in MT VMs).
 */
void
	ltl_threadenv_activate(
		ltl_threadenv_t *
	);

/*
 * Marks a thread environment as inactive (allows proper exception
 * handling - does nothing in MT VMs).
 */
void
	ltl_threadenv_deactivate(
		ltl_threadenv_t *
	);

#endif



/*
 * Returns true, if this threadenv is currently active.
 */
int
	ltl_threadenv_active(
		ltl_threadenv_t *
	);

/*
 * Returns the currently active VM thread execution context, if any.
 */
ltl_threadenv_t *
	ltl_threadenv_current();

/*
 * Dispose the thread environment.
 */
void
	ltl_threadenv_dispose(
		ltl_threadenv_t *
	);

/*
 * Returns the ltl_env structure for the given threadenv.
 */
ltl_env_t *
	ltl_threadenv_getenv(
		ltl_threadenv_t *
	);

/*
 * Convenience: passes call to ltl_env_getsym().
 */
ltl_sym_t
	ltl_threadenv_getsym(
		ltl_threadenv_t *,
		const char *
	);

/*
 * Returns the global binding of a symbol - if the symbol represents a class,
 * it is dynamically loaded into memory, if this hasn't yet happened. The function
 * may rise an exception, if the dynamic load fails, or if the symbol isn't known
 * at all (symbols have to be registered via ltl_env_getsym() first).
 */
ltl_object_t *
	ltl_threadenv_get_global4sym(
		ltl_threadenv_t *,
		ltl_sym_t
	);

/*
 * Returns the null-terminated string represented by a given symbol token
 * An exception is raised, if the symbol is unknown (symbols have to be registered
 * via ltl_env_getsym() first).
 */
char *
	ltl_threadenv_get_str4sym(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_sym_t		/* A symbol.			*/
	);

/*
 * Returns the numeric constant referred by handle.
 */
ltl_word_t
	ltl_threadenv_get_const4ref(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_key_t			/* Constant handle.			*/
	);

/*
 * Returns the string constant referred by handle.
 */
char *
	ltl_threadenv_get_psz4ref(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_key_t			/* Constant handle.			*/
	);

/*
 * Finds a method implementation for a given type and feature symbol.
 * Raises an exception, if no implementation is found.
 */
ltl_object_t *
	ltl_threadenv_get_method4tf(
		ltl_threadenv_t *,	/* Little environment.	*/
		ltl_sym_t,		/* Type symbol.		*/
		ltl_sym_t		/* Feature symbol.	*/
	);
/*
 * Finds an attribute descriptor for a given type and feature symbol.
 * Raises an exception, if no implementation is found.
 */
ltl_object_t * 
	ltl_threadenv_get_attrdesc4tf(
		ltl_threadenv_t *,	/* Little environment.	*/
		ltl_sym_t,		/* Asking type symbol.	*/
		ltl_sym_t,		/* Owner type symbol.	*/
		ltl_sym_t		/* Feature symbol.	*/
	);

/*
 * Pushes a call argument onto the execution stack.
 * Returns the pointer to the argument's stack location.
 */
ltl_object_t *
	ltl_threadenv_pusharg(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_object_t *		/* Call argument.		*/
	);

/*
 * Pushes the stack pointer and returns it.
 */
ltl_object_t *
	ltl_threadenv_push(
		ltl_threadenv_t *	/* Little thread context.	*/
	);

/*
 * Invoke a method, whose stack frame has been previously built using
 * the ltl_threadenv_pusharg() routine, and whose method object has already
 * been retrieved.
 * This is the fast, but inconvenient way to invoke a method.
 */
ltl_object_t *
	ltl_threadenv_invoke_with_frame(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_object_t *,		/* Calling object.		*/
		ltl_object_t *		/* Method object.		*/
	);

/*
 * Convenient way of invoking methods.
 * This function builds the threadenv stack frame from the passed parameter
 * list (which must consist of pointers to ltl_object_t).
 */
ltl_object_t *
	ltl_threadenv_invoke(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_object_t *,		/* Calling object.		*/
		ltl_sym_t,		/* Method symbol.		*/
		int,			/* Number of arguments.		*/
		...			/* List of arguments.		*/
	);

/*
 * Returns the result of the previous computation (the item
 * on top of the VM stack.
 */
ltl_object_t *
	ltl_threadenv_getresult(
		ltl_threadenv_t *	/* Little thread context.	*/
	);

/*
 * Returns an eventually pending exception.
 * If none, LTL_NIL will be returned.
 */
ltl_object_t *
	ltl_threadenv_getexception(
		ltl_threadenv_t *	/* Little thread context.	*/
	);

/*
 * Clears exception info of previous errors in this thread context. 
 */
void
	ltl_threadenv_clearexception(
		ltl_threadenv_t *	/* Little thread context.	*/
	);

/*
 * Used to raise VM-detected exceptions (see error ID definitions).
 */
void
	ltl_threadenv_error_2(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_sym_t		/* Error symbol.		*/
	);
void
	ltl_threadenv_error_3(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_sym_t,		/* Error symbol.		*/
		char *			/* Message.			*/
	);
void
	ltl_threadenv_error_4(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_sym_t,		/* Error symbol			*/
		char *,			/* Message 1.			*/
		char *			/* Message 2.			*/
	);
        
/*
 * Pushes a jump buffer for setjmp()/longjmp() based exception
 * handling.
 */
void *
	_ltl_threadenv_push_xframe(
		ltl_threadenv_t *,	/* Little thread context. 	*/
		ltl_object_t *,		/* Method to save.		*/
		int,			/* Label to save.		*/
		ltl_object_t *,		/* Stack pointer to save.	*/
		ltl_object_t *,		/* Frame pointer to save.	*/
		char *,
		int
	);

#define ltl_threadenv_push_xframe(t,m,l,sp,fp) \
 _ltl_threadenv_push_xframe((t),(m),(l),(sp),(fp),__FILE__,__LINE__)

/*
 * Pushes a jump buffer for exception handling.
 * No VM environment saved.
 */
void *
	_ltl_threadenv_push_xframe_noenv(
		ltl_threadenv_t *,
		char *,
		int
	);

#define ltl_threadenv_push_xframe_noenv(t) \
 _ltl_threadenv_push_xframe_noenv((t),__FILE__,__LINE__)

/*
 * Removes an exception handler from the VM's handler stack.
 * In a supervised area, each exit point must be guarded by calls to
 * ltl_threadenv_pop_xframe() to restore the environment.
 */
void
        ltl_threadenv_pop_xframe(
                ltl_threadenv_t *	/* Little thread context.	*/
        );
        
	
/*
 * Throw an exception (any object - API users should make sure that
 * the application knows how to handle the error).
 * However, if the ltl_env_exec_entity() function is used to execute
 * a class' main method, generic exception handling is done and correct
 * termination granted.
 */
void
	ltl_threadenv_throw(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_object_t *		/* Exception object.		*/
	);
	
/*
 * Rethrows the object currently on stack top.
 */
void
	ltl_threadenv_rethrow(
		ltl_threadenv_t *	/* Little thread context.	*/
	);

/*
 * Throws an uncatchable error, which leads to immediate termination
 * of the current VM execution thread.
 */
void
	ltl_threadenv_fatal(
		ltl_threadenv_t *,	/* Little thread context.	*/
		ltl_object_t *		/* Exception object.		*/
	);

/*
 * This is how Little native methods should be declared:
 */
#define LTL_METHOD_DECL(name) \
	ltl_object_t * name (\
		ltl_threadenv_t *threadenv,\
		ltl_object_t *method,\
		ltl_object_t *self,\
		ltl_object_t *argv)

/*
 * Macros for exception handling:
 *
 * LTL_TRY_BEGIN(threadenv) {
 * 	...
 * }
 * LTL_CATCH(threadenv,exception) {
 *	...
 * }
 * LTL_TRY_END(threadenv);
 *
 * Important note:
 * do never explicitly exit a try- or catch-block - this
 * will prevent the ALICE exception handler to be removed
 * and lead to errative behaviour on future exceptions!
 */                
#define LTL_TRY_BEGIN(threadenv) {\
	void *buf = NULL;\
	ltl_threadenv_activate(threadenv);\
	buf = ltl_threadenv_push_xframe_noenv(threadenv);\
	if ( ! sigsetjmp((ltl_jmp_buf)buf,1) ) {
        
#define LTL_CATCH(threadenv) \
		ltl_threadenv_pop_xframe(threadenv);\
		ltl_threadenv_deactivate(threadenv);\
	}\
	else {\
		ltl_object_t *exception = ltl_threadenv_getexception(threadenv);\
		ltl_threadenv_pop_xframe(threadenv);\
		ltl_threadenv_deactivate(threadenv);
        
#define LTL_TRY_END(threadenv) \
		if ( threadenv ) {\
			ltl_threadenv_clearexception(threadenv);\
		}\
	}}

#define LTL_ARG(n) (argv+(n))
#define LTL_ARGC() argc

#define LTL_LOCAL() ltl_threadenv_push(threadenv)

#define LTL_RETURN(o) return o
#define LTL_RETURN_SPECIAL(t,v) \
{ ltl_object_t *rv = LTL_LOCAL(); LTL_OBJECT_SETSPECIAL(rv,t,v); return rv; }

#ifdef __cplusplus
}
#endif

#endif
